@@ -9,7 +9,6 @@ from datetime import datetime
 from glob import glob
 from pathlib import Path
 from subprocess import CalledProcessError
-from tempfile import TemporaryDirectory
 
 import requests
 from ghapi.core import GhApi
@@ -35,7 +34,7 @@ def bump_version(version_spec, version_cmd):
 
     # Bail if tag already exists
     tag_name = f"v{version}"
-    if tag_name in util.run("git --no-pager tag").splitlines():
+    if tag_name in util.run("git --no-pager tag", quiet=True).splitlines():
         msg = f"Tag {tag_name} already exists!"
         msg += " To delete run: `git push --delete origin {tag_name}`"
         raise ValueError(msg)
@@ -49,6 +48,7 @@ def check_links(ignore_glob, ignore_links, cache_file, links_expire):
     os.makedirs(cache_dir, exist_ok=True)
     cmd = "pytest --noconftest --check-links --check-links-cache "
     cmd += f"--check-links-cache-expire-after {links_expire} "
+    cmd += "--disable-warnings --quiet "
     cmd += f"--check-links-cache-name {cache_dir}/check-release-links "
 
     ignored = []
@@ -76,20 +76,27 @@ def check_links(ignore_glob, ignore_links, cache_file, links_expire):
     for f in files:
         file_cmd = cmd + f' "{f}"'
         try:
-            util.run(file_cmd)
+            util.run(file_cmd, shell=False)
         except Exception as e:
             # Return code 5 means no tests were run (no links found)
             if e.returncode != 5:
-                util.run(file_cmd + " --lf")
+                util.run(file_cmd + " --lf", shell=False)
 
 
-def draft_changelog(version_spec, branch, repo, since, auth, changelog_path, dry_run):
+def draft_changelog(
+    version_spec, branch, repo, since, since_last_stable, auth, changelog_path, dry_run
+):
     """Create a changelog entry PR"""
     repo = repo or util.get_repo()
     branch = branch or util.get_branch()
     version = util.get_version()
 
-    tags = util.run("git --no-pager tag")
+    # Check for multiple versions
+    npm_versions = None
+    if util.PACKAGE_JSON.exists():
+        npm_versions = npm.get_package_versions(version)
+
+    tags = util.run("git --no-pager tag", quiet=True)
     if f"v{version}" in tags.splitlines():
         raise ValueError(f"Tag v{version} already exists")
 
@@ -108,11 +115,10 @@ def draft_changelog(version_spec, branch, repo, since, auth, changelog_path, dry
     commit_message = f'git commit -a -m "{title}"'
     body = title
 
-    # Check for multiple versions
-    if util.PACKAGE_JSON.exists():
-        body += npm.get_package_versions(version)
+    if npm_versions:
+        body += npm_versions
 
-    body += '\n\nAfter merging this PR run the "Draft Release" Workflow on your fork of `jupyter_releaser` with the following inputs'
+    body += '\n\nAfter merging this PR run the "Full Release" Workflow on your fork of `jupyter_releaser` with the following inputs'
     body += f"""
 | Input  | Value |
 | ------------- | ------------- |
@@ -120,7 +126,9 @@ def draft_changelog(version_spec, branch, repo, since, auth, changelog_path, dry
 | Branch  | {branch}  |
 | Version Spec | {version_spec} |
 """
-    if since:
+    if since_last_stable:
+        body += "| Since Last Stable | true |"
+    elif since:
         body += f"| Since | {since} |"
     util.log(body)
 
@@ -169,19 +177,22 @@ def make_changelog_pr(auth, branch, repo, title, commit_message, body, dry_run=F
     util.actions_output("pr_url", pull.html_url)
 
 
-def tag_release(dist_dir, no_git_tag_workspace):
+def tag_release(
+    dist_dir, release_message, tag_format, tag_message, no_git_tag_workspace
+):
     """Create release commit and tag"""
     # Get the new version
     version = util.get_version()
 
     # Create the release commit
-    util.create_release_commit(version, dist_dir)
+    util.create_release_commit(version, release_message, dist_dir)
 
     # Create the annotated release tag
-    tag_name = f"v{version}"
-    util.run(f'git tag {tag_name} -a -m "Release {tag_name}"')
+    tag_name = tag_format.format(version=version)
+    tag_message = tag_message.format(tag_name=tag_name, version=version)
+    util.run(f'git tag {tag_name} -a -m "{tag_message}"')
 
-    # Create annotated release tags for workspace packages if given
+    # Create release tags for workspace packages if given
     if not no_git_tag_workspace:
         npm.tag_workspace_packages()
 
@@ -196,6 +207,7 @@ def draft_release(
     dist_dir,
     dry_run,
     post_version_spec,
+    post_version_message,
     assets,
 ):
     """Publish Draft GitHub release and handle post version bump"""
@@ -209,8 +221,7 @@ def draft_release(
     # Bump to post version if given
     if post_version_spec:
         post_version = bump_version(post_version_spec, version_cmd)
-
-        util.log(f"Bumped version to {post_version}")
+        util.log(post_version_message.format(post_version=post_version))
         util.run(f'git commit -a -m "Bump to {post_version}"')
 
     if dry_run:
@@ -239,7 +250,7 @@ def draft_release(
     release = gh.create_release(
         f"v{version}",
         branch,
-        f"Release v{version}",
+        f"v{version}",
         body,
         True,
         prerelease,
@@ -271,8 +282,20 @@ def extract_release(auth, dist_dir, dry_run, release_url, npm_install_options):
     owner, repo = match["owner"], match["repo"]
     gh = GhApi(owner=owner, repo=repo, token=auth)
     release = util.release_for_url(gh, release_url)
+    branch = release.target_commitish
     assets = release.assets
 
+    # Prepare a git checkout
+    prep_git(None, branch, f"{owner}/{repo}", auth, None, None)
+
+    orig_dir = os.getcwd()
+    os.chdir(util.CHECKOUT_NAME)
+
+    # Read in the config
+    config = util.read_config()
+    options = config.get("options", {})
+    npm_install_options = npm_install_options or options.get("npm-install-options", "")
+
     # Clean the dist folder
     dist = Path(dist_dir)
     if dist.exists():
@@ -290,19 +313,24 @@ def extract_release(auth, dist_dir, dry_run, release_url, npm_install_options):
             with open(path, "wb") as f:
                 for chunk in r.iter_content(chunk_size=8192):
                     f.write(chunk)
-            suffix = Path(asset.name).suffix
-            if suffix in [".gz", ".whl"]:
-                python.check_dist(path)
-            elif suffix == ".tgz":
-                npm.check_dist(path, npm_install_options)
-            else:
-                util.log(f"Nothing to check for {asset.name}")
+
+    # Check all npm packages
+    npm.check_dist(dist, npm_install_options)
+
+    # Check python packages individually
+    for asset in assets:
+        suffix = Path(asset.name).suffix
+        if suffix in [".gz", ".whl"]:
+            python.check_dist(dist / asset.name)
+        elif suffix == ".tgz":
+            pass  # already handled
+        else:
+            util.log(f"Nothing to check for {asset.name}")
 
     # Skip sha validation for dry runs since the remote tag will not exist
     if dry_run:
         return
 
-    branch = release.target_commitish
     tag_name = release.tag_name
 
     sha = None
@@ -312,17 +340,9 @@ def extract_release(auth, dist_dir, dry_run, release_url, npm_install_options):
     if sha is None:
         raise ValueError("Could not find tag")
 
-    # Run a git checkout
-    # Fetch the branch
-    # Get the commmit message for the branch
+    # Get the commmit message for the tag
     commit_message = ""
-    with TemporaryDirectory() as td:
-        url = gh.repos.get().html_url
-        util.run(f"git clone {url} local", cwd=td)
-        checkout = osp.join(td, "local")
-        if not osp.exists(url):
-            util.run(f"git fetch origin {branch}", cwd=checkout)
-        commit_message = util.run(f"git log --format=%B -n 1 {sha}", cwd=checkout)
+    commit_message = util.run(f"git log --format=%B -n 1 {sha}")
 
     for asset in assets:
         # Check the sha against the published sha
@@ -340,6 +360,8 @@ def extract_release(auth, dist_dir, dry_run, release_url, npm_install_options):
         if not valid:  # pragma: no cover
             raise ValueError(f"Invalid file {asset.name}")
 
+    os.chdir(orig_dir)
+
 
 def parse_release_url(release_url):
     """Parse a release url into a regex match"""
@@ -350,19 +372,19 @@ def parse_release_url(release_url):
     return match
 
 
-def publish_assets(dist_dir, npm_token, npm_cmd, twine_cmd, dry_run, use_checkout_dir):
+def publish_assets(
+    dist_dir, npm_token, npm_cmd, twine_cmd, npm_registry, twine_registry, dry_run
+):
     """Publish release asset(s)"""
-    if use_checkout_dir:
-        if not osp.exists(util.CHECKOUT_NAME):
-            raise ValueError("Please run prep-git first")
-        os.chdir(util.CHECKOUT_NAME)
+    os.environ["NPM_REGISTRY"] = npm_registry
+    os.environ["TWINE_REGISTRY"] = twine_registry
 
     if dry_run:
         # Start local pypi server with no auth, allowing overwrites,
         # in a temporary directory
         if len(glob(f"{dist_dir}/*.whl")):
             python.start_local_pypi()
-            twine_cmd = "twine upload --repository-url=http://localhost:8081"
+            twine_cmd = "twine upload --repository-url=http://0.0.0.0:8081"
             os.environ["TWINE_USERNAME"] = "foo"
             os.environ["TWINE_PASSWORD"] = "bar"
         npm_cmd = "npm publish --dry-run"
@@ -370,19 +392,23 @@ def publish_assets(dist_dir, npm_token, npm_cmd, twine_cmd, dry_run, use_checkou
         os.environ.setdefault("TWINE_USERNAME", "__token__")
 
     if len(glob(f"{dist_dir}/*.tgz")):
-        npm.handle_npm_config(npm_token, dist_dir)
+        npm.handle_npm_config(npm_token)
+        if npm_token:
+            util.run("npm whoami")
 
     found = False
     for path in sorted(glob(f"{dist_dir}/*.*")):
         name = Path(path).name
         suffix = Path(path).suffix
         if suffix in [".gz", ".whl"]:
-            util.run(f"{twine_cmd} {name}", cwd=dist_dir)
+            util.retry(f"{twine_cmd} {name}", cwd=dist_dir)
             found = True
         elif suffix == ".tgz":
             # Ignore already published versions
             try:
-                util.run(f"{npm_cmd} {name}", cwd=dist_dir, quiet=True)
+                util.run(
+                    f"{npm_cmd} {name}", cwd=dist_dir, quiet=True, quiet_error=True
+                )
             except CalledProcessError as e:
                 stderr = e.stderr
                 if (
@@ -423,14 +449,14 @@ def publish_release(auth, release_url):
     util.actions_output("release_url", release.html_url)
 
 
-def prep_git(ref, branch, repo, auth, username, url, install=True):
+def prep_git(ref, branch, repo, auth, username, url):
     """Set up git"""
     repo = repo or util.get_repo()
 
     user_name = ""
     try:
         user_name = util.run("git config --global user.email")
-    except Exception as e:
+    except Exception:
         pass
 
     if not user_name:
@@ -470,7 +496,7 @@ def prep_git(ref, branch, repo, auth, username, url, install=True):
     ref = ref or ""
 
     # Make sure we have *all* tags
-    util.run("git fetch origin --tags --force")
+    util.run("git fetch origin --tags --force --quiet")
 
     # Handle the ref
     if ref.startswith("refs/pull/"):
@@ -496,21 +522,26 @@ def prep_git(ref, branch, repo, auth, username, url, install=True):
     else:
         util.run(checkout_cmd)
 
+    # Check for detached head state, create branch if needed
+    try:
+        util.run("git symbolic-ref -q HEAD")
+    except Exception:
+        util.run(f"git switch -c {branch}")
+
     # Install the package
-    if install:
-        # install python package in editable mode with test deps
-        if util.SETUP_PY.exists():
-            util.run('pip install -e ".[test]"')
+    # install python package in editable mode with test deps
+    if util.SETUP_PY.exists():
+        util.run('pip install -q -e ".[test]"')
 
-        # prefer yarn if yarn lock exists
-        elif util.YARN_LOCK.exists():
-            if not shutil.which("yarn"):
-                util.run("npm install -g yarn")
-            util.run("yarn")
+    # prefer yarn if yarn lock exists
+    elif util.YARN_LOCK.exists():
+        if not shutil.which("yarn"):
+            util.run("npm install -g yarn")
+        util.run("yarn")
 
-        # npm install otherwise
-        elif util.PACKAGE_JSON.exists():
-            util.run("npm install")
+    # npm install otherwise
+    elif util.PACKAGE_JSON.exists():
+        util.run("npm install")
 
     os.chdir(orig_dir)
 
@@ -518,7 +549,7 @@ def prep_git(ref, branch, repo, auth, username, url, install=True):
 
 
 def forwardport_changelog(
-    auth, ref, branch, repo, username, changelog_path, dry_run, git_url, release_url
+    auth, ref, branch, repo, username, changelog_path, dry_run, release_url
 ):
     """Forwardport Changelog Entries to the Default Branch"""
     # Set up the git repo with the branch
@@ -526,16 +557,17 @@ def forwardport_changelog(
     gh = GhApi(owner=match["owner"], repo=match["repo"], token=auth)
     release = util.release_for_url(gh, release_url)
     tag = release.tag_name
+    source_branch = release.target_commitish
 
     repo = f'{match["owner"]}/{match["repo"]}'
 
-    # We want to target the main branch
-    orig_dir = os.getcwd()
-    branch = prep_git(None, None, repo, auth, username, git_url, install=False)
-    os.chdir(util.CHECKOUT_NAME)
+    # switch to main branch here
+    branch = branch or util.get_default_branch()
+    util.run(f"git fetch origin {branch}")
+    util.run(f"git checkout {branch}")
 
     # Bail if the tag has been merged to the branch
-    tags = util.run(f"git --no-pager tag --merged {branch}")
+    tags = util.run(f"git --no-pager tag --merged {branch}", quiet=True)
     if tag in tags.splitlines():
         util.log(f"Skipping since tag is already merged into {branch}")
         return
@@ -593,5 +625,4 @@ def forwardport_changelog(
     )
 
     # Clean up after ourselves
-    os.chdir(orig_dir)
-    shutil.rmtree(util.CHECKOUT_NAME)
+    util.run(f"git checkout {source_branch}")
